function ydot = yderivs_3D(t,y);
global g gam m R p23 p34 pc3 pc4 p4f Icm3 Icm4 h hspr hdamp sspr sdamp

% Code by Mariano Garcia (msg5@cornell.edu), 1998

% 3D file for derivatives of 2-link walker, general mass distribution
% and with point-feet.
% Modified for disc feet 3/18/98 and checked against Greenwood
% using init_params.m
% Disk feet do not destroy symmetry of M.

% hfoot1 and hdotfoot1 are the z1-height and z1-speed
% of the [bottom of the] swing foot in frame 1.

% define angles and rotation matrices (see my thesis)

cO1 = cos(y(1));
sO1 = sin(y(1));

cO2 = cos(y(2));
sO2 = sin(y(2));

cO3 = cos(y(3));
sO3 = sin(y(3));

sO4 = sin(y(4));
cO4 = cos(y(4));


R01=[cO1 -sO1 0 ;sO1 cO1 0; 0 0 1];
R12=[cO2 -sO2 0 ;0 0 -1;sO2 cO2 0];
R23=[cO3 -sO3 0 ;0 0 -1;sO3 cO3 0];
R34=[cO4 -sO4 0 ;sO4 cO4 0; 0 0 1];

R10=R01';
R21=R12';
R32=R23';
R43=R34';

% The derivative file is based on a recursive method to get
% the four state angle accelerations as functions of the state.
% More detail is given in my thesis, especially chapter 2.
% The idea is that the equations of motion can be put in the form
% tau = M*O'' + V + G where
% tau = 4 x 1 vector of applied joint torques
% M = 4 x 4 mass matrix, also used for heelstrike
% O'' =  4 x 1 vector of joint angle accelerations (unknown)
% V = 4 x 1 vector of coriolis and centrifugal terms
% G = 4 x 1 vector of gravity terms.
% We can get the above terms one-at-a-time and then solve
% numerically for O''.
% The swing leg is not contacting anything so there are no
% torques or forces at the free end during free-swinging.

% first construct [V+G] and tau. Assume temporarily that
% O'' = [0 0 0 0]'; "Accelerations" here are partial accelerations.

vg=zeros(4,1);

vdot0 = [0 g*sin(gam) g*cos(gam)]'; % accelerate the base frame
                                    % upwards as a substitute for
                                    % putting a gravity force on
                                    % every link.

vdot1 = R10*vdot0; % the acceleration of frame 1 with O'' = 0

omega1=[0 0 y(5)]'; % omegas are the absolute angular velocities
                    % of the corresponding frames.

omega2=R21*omega1+[0 0 y(6)]';
omega2dot=scross(R21*omega1,[0 0 y(6)]');

omega3=R32*omega2+[0 0 y(7)]'; % this is the angular velocity of the
                               % stance leg (frame 3)

omega3dot=R32*omega2dot+scross(R32*omega2,[0 0 y(7)]'); % partial accel
                               % of the stance leg (frame 3)

p233=R32*p23;  % vector from contact point to center of stance foot
               % but written in frame 2 coords

v3 = scross(omega3,p233); %vel of stance foot center

   % kludge for rolling contact. 
vdot3an= [sO2*y(6)*y(5)*sO3*R+(-cO2*y(5)+y(7))*cO3*y(7)*R
        sO2*y(6)*y(5)*cO3*R-(-cO2*y(5)+y(7))*sO3*y(7)*R
        0 ];  % from omegadot x r but with thetaddots =0

% the line below is the partial accel of stance foot center in frame 3
vdot3 = vdot3an + scross(omega3,v3)+ R32*R21*vdot1; 
% the line below is the partial accel of stance leg cm in frame 3
vcdot3 = scross(omega3dot,pc3)+scross(omega3,scross(omega3,pc3))+vdot3;


F3 = m(3)*vcdot3;  % partial "inertial force" of stance leg. I don't like
                   % this terminology but it follows from the algorithm
                   % that I use.
N3 = Icm3*omega3dot+scross(omega3,Icm3*omega3); % partial "inertial torque"
                   % of stance leg.

omega4=R43*omega3+[0 0 y(8)]'; % the angular velocity of the swing leg (frame 4)
omega4dot=R43*omega3dot+scross(R43*omega3,[0 0 y(8)]'); % the partial
                               % angular accel of the swing leg (frame 4)

% the line below is the partial accel of swing foot center in frame 4
vdot4 = R43*(scross(omega3dot,p34)+scross(omega3,scross(omega3,p34))+vdot3);
% the line below is the partial accel of swing leg cm in frame 4
vcdot4 = scross(omega4dot,pc4)+scross(omega4,scross(omega4,pc4))+vdot4;

% for collision detection, calculate the height and vertical velocity
% of the swing foot bottom.
vf3 =v3+scross(omega3,p34)+R34*(scross(omega4,(p4f-R43*p233))); 
hfdot1=R12*(R23*vf3);  % swing foot bottom velocity
hfdot1=hfdot1(3);  % get the vertical component only
hf1=R12*(R23*(p233+p34+R34*(p4f-R43*p233))); % swing foot position
hf1=hf1(3);  % get the height only

F4 = m(4)*vcdot4; % partial "inertial force" of swing leg
N4 = Icm4*omega4dot+scross(omega4,Icm4*omega4); % partial "inertial torque"
                   % of stance leg.

% Now do "f=ma" and "n=Iw+..."

f4 = F4; %"f = ma" for link 4
n4 = N4+scross(pc4,F4); % "n = Iw" for link 4
vg(4) = n4(3); % x4 and y4 directions are constrained, so the
               % only rotation is in the z4 direction

f3 = R34*f4+F3; %"f = ma" for link 3
n3 = N3+R34*n4+scross(pc3,F3)+scross(p34,R34*f4)-scross(-p233,f3); %"n = Iw" link 3
vg(3) = n3(3); % x3 and y3 irections are constrained, so the
               % only rotation is in the z3 direction.

% same deal for links 2 and 1 but they have no mass or inertia associated with em
n2 = R23*n3;%+scross(p23,R23*f3);
vg(2) = n2(3);

n1 = R12*n2;
vg(1) = n1(3);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%     Now get mass matrix column by column         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% construct column 1 by taking O1dotdot=1, all other accels and  velocs = 0

M=zeros(4);

omega1dot = [0 0 1]';  % the (assumed temporary) angular acceleration

omega2dot=R21*omega1dot; % other angular accels are just a rotation
                         % of this one into a new frame
omega3dot=R32*omega2dot;

vdot3 = scross(omega3dot,p233);  % partial accel of frame 3
vcdot3 = scross(omega3dot,pc3)+vdot3; % partial accel of cm 3

F3 = m(3)*vcdot3;  % partial "inertial force" of stance leg
N3 = Icm3*omega3dot; % partial "inertial torque" of stance leg

omega4dot=R43*omega3dot;

vdot4 = R43*(scross(omega3dot,p34)+vdot3); %partial accel of frame 3 (swing)
vcdot4 = scross(omega4dot,pc4)+vdot4; % partial accel of cm 4 (swing leg)

F4 = m(4)*vcdot4; % partial "inertial force" of swing leg
N4 = Icm4*omega4dot; % partial "inertial torque" of swing leg

% Now do "f=ma" and "n=Iw+..." for all links as before

f4 = F4;
n4 = N4+scross(pc4,F4);
M(4,1) = n4(3);

f3 = R34*f4+F3;
n3 = N3+R34*n4+scross(pc3,F3)+scross(p34,R34*f4)-scross(-p233,f3);
M(3,1) = n3(3);

f2 = R23*f3;
n2 = R23*n3;%+scross(p23,R23*f3);
M(2,1) = n2(3);

n1 = R12*n2;
M(1,1) = n1(3);


%%%%%%%%%%%%%%%%%%%%%%%%%%%
% construct column 2  in the same way but with thetadoubledot(2) = 1
% and all other thetadoubledots and thetadots = 0

omega2dot=[0 0 1]';

omega3dot=R32*omega2dot;

vdot3 = scross(omega3dot, p233);
vcdot3 = scross(omega3dot,pc3)+vdot3;

F3 = m(3)*vcdot3;
N3 = Icm3*omega3dot;

omega4dot=R43*omega3dot;

vdot4 = R43*(scross(omega3dot,p34)+vdot3);
vcdot4 = scross(omega4dot,pc4)+vdot4;

F4 = m(4)*vcdot4;
N4 = Icm4*omega4dot;

f4 = F4;
n4 = N4+scross(pc4,F4);
M(4,2) = n4(3);

f3 = R34*f4+F3;
n3 = N3+R34*n4+scross(pc3,F3)+scross(p34,R34*f4)-scross(-p233,f3);
M(3,2) = n3(3);

n2 = R23*n3;%+scross(p23,R23*f3);
M(2,2) = n2(3);

M(1,2) = M(2,1);  % use the symmetry of M to save some flops


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% construct column 3

omega3dot=[0 0 1]';

vdot3 = scross(omega3dot,p233);
vcdot3 = scross(omega3dot,pc3)+vdot3;

F3 = m(3)*vcdot3;
N3 = Icm3*omega3dot;

omega4dot=R43*omega3dot;

vdot4 = R43*(scross(omega3dot,p34)+vdot3);
vcdot4 = scross(omega4dot,pc4)+vdot4;

F4 = m(4)*vcdot4;
N4 = Icm4*omega4dot;

f4 = F4;
n4 = N4+scross(pc4,F4);
M(4,3) = n4(3);

f3 = R34*f4+F3;
n3 = N3+R34*n4+scross(pc3,F3)+scross(p34,R34*f4)-scross(-p233,f3);
M(3,3) = n3(3);

M(2,3) = M(3,2);

M(1,3) = M(3,1);



%%%%%%%%%%%%%%%%%%%%%%%%%%
% construct column 4 % careful w/ copy/pasting non-zero F3 and N3 here

omega4dot=[0 0 1]';

vcdot4 = scross(omega4dot,pc4);

F4 = m(4)*vcdot4;
N4 = Icm4*omega4dot;

f4 = F4;
n4 = N4+scross(pc4,F4);
M(4,4) = n4(3);

M(3,4) = M(4,3);

M(2,4) = M(4,2);

M(1,4) =M(4,1);

%M=M(1:3,1:3);
%vg=vg(1:3);

% next line is torques due to springs and dampers
tau=[-sspr*y(1)-sdamp*y(5) 0 0 -hspr*(y(4)-pi)-hdamp*y(8)]';

% solve for O''
X=chol(M);
z = X'\(tau-vg);
Oddot=X\z;

%Oddot=M\-vg;

vpdot = R01*[0 -y(7)*R 0]'; % velocity of contact point, used for graphics

% construct the returning vector depending on the size of the incoming
% vector. This tells us if we are doing forward integration with
% graphics (int_walker_3D.m) or gait fixed-point finding (it_3D_mcg.m).

if length(y) == 8 % gait fixed-point finding - no need for graphics info
ydot=[y(5) y(6) y(7) y(8) Oddot(1) Oddot(2) Oddot(3) Oddot(4) hf1 hfdot1]';
end;

if length(y) == 10 % forward integration with graphics
ydot=[y(5) y(6) y(7) y(8) Oddot(1) Oddot(2) Oddot(3)  Oddot(4) ...
     vpdot(1) vpdot(2) hf1 hfdot1 ]'; % longer one for int
end;

% leftovers from debugging - occasionally convenient
% ydot=[y(5) y(6) y(7) y(8) Oddot(1) Oddot(2) Oddot(3) Oddot(4)]'
% pause(500);
